@@ -103,30 +103,6 @@ size_t sumup_dir( const string &dir )
     return res;
 }
 
-static string list_native_environment( const string &nativedir )
-{
-    assert( nativedir.at( nativedir.length() - 1 ) == '/' );
-
-    string native_environment;
-
-    DIR *tdir = opendir( nativedir.c_str() );
-    if ( tdir ) {
-        string suff = ".tar.gz";
-        do {
-            struct dirent *myenv = readdir(tdir);
-            if ( !myenv )
-                break;
-            string versfile = myenv->d_name;
-            if ( versfile.size() > suff.size() && versfile.substr( versfile.size() - suff.size() ) == suff ) {
-                native_environment = nativedir + versfile;
-                break;
-            }
-        } while ( true );
-        closedir( tdir );
-    }
-    return native_environment;
-}
-
 static void list_target_dirs( const string &current_target, const string &targetdir, Environments &envs )
 {
     DIR *envdir = opendir( targetdir.c_str() );
@@ -136,7 +112,7 @@ static void list_target_dirs( const string &current_target, const string &target
     for ( struct dirent *ent = readdir(envdir); ent; ent = readdir( envdir ) )
     {
         string dirname = ent->d_name;
-        if ( !access( string( targetdir + "/" + dirname + "/usr/bin/gcc" ).c_str(), X_OK ) )
+        if ( !access( string( targetdir + "/" + dirname + "/usr/bin/as" ).c_str(), X_OK ) )
             envs.push_back( make_pair( current_target, dirname ) );
     }
     closedir( envdir );
@@ -209,64 +185,135 @@ Environments available_environmnents(const string &basedir)
     return envs;
 }
 
-size_t setup_env_cache(const string &basedir, string &native_environment, uid_t nobody_uid, gid_t nobody_gid)
+void save_compiler_timestamps(time_t &gcc_bin_timestamp, time_t &gpp_bin_timestamp, time_t &clang_bin_timestamp)
+{
+    struct stat st;
+    if( stat( "/usr/bin/gcc", &st ) == 0 )
+        gcc_bin_timestamp = st.st_mtime;
+    else
+        gcc_bin_timestamp = 0;
+    if( stat( "/usr/bin/g++", &st ) == 0 )
+        gpp_bin_timestamp = st.st_mtime;
+    else
+        gpp_bin_timestamp = 0;
+    if( stat( "/usr/bin/clang", &st ) == 0 )
+        clang_bin_timestamp = st.st_mtime;
+    else
+        clang_bin_timestamp = 0;
+}
+
+bool compilers_uptodate(time_t gcc_bin_timestamp, time_t gpp_bin_timestamp, time_t clang_bin_timestamp)
+{
+    struct stat st;
+    if( stat( "/usr/bin/gcc", &st ) == 0 ) {
+        if( st.st_mtime != gcc_bin_timestamp )
+            return false;
+    } else {
+        if( gcc_bin_timestamp != 0 )
+            return false;
+    }
+    if( stat( "/usr/bin/g++", &st ) == 0 ) {
+        if( st.st_mtime != gpp_bin_timestamp )
+            return false;
+    } else {
+        if( gpp_bin_timestamp != 0 )
+            return false;
+    }
+    if( stat( "/usr/bin/clang", &st ) == 0 ) {
+        if( st.st_mtime != clang_bin_timestamp )
+            return false;
+    } else {
+        if( clang_bin_timestamp != 0 )
+            return false;
+    }
+    return true;
+}
+
+size_t setup_env_cache(const string &basedir, string &native_environment, uid_t user_uid, gid_t user_gid,
+                       const std::string &compiler, const list<string> &extrafiles)
 {
     native_environment = "";
     string nativedir = basedir + "/native/";
 
-    if ( ::access( "/usr/bin/gcc", X_OK ) || ::access( "/usr/bin/g++", X_OK ) ) 
-	return 0;
+    if (compiler == "clang") {
+        if ( ::access( "/usr/bin/clang", X_OK ) != 0 )
+            return 0;
+    } else { // "gcc" (the default)
+        // Both gcc and g++ are needed in the gcc case.
+        if ( ::access( "/usr/bin/gcc", X_OK ) != 0 || ::access( "/usr/bin/g++", X_OK ) != 0 )
+	    return 0;
+    }
 
-    if ( mkdir( nativedir.c_str(), 0775 ) )
+    if ( mkdir( nativedir.c_str(), 0775 ) && errno != EEXIST )
    	return 0; 
 
-    if ( chown( nativedir.c_str(), 0, nobody_gid ) ||
+    if ( chown( nativedir.c_str(), 0, user_gid ) ||
          chmod( nativedir.c_str(), 0775 ) ) {
 	rmdir( nativedir.c_str() );
 	return 0;
     }
 
     flush_debug();
+    int pipes[ 2 ];
+    pipe( pipes );
     pid_t pid = fork();
     if ( pid ) {
+        close( pipes[ 1 ] );
         int status = 1;
         while ( waitpid( pid, &status, 0 ) < 0 && errno == EINTR )
            ;
         trace() << "waitpid " << status << endl;
         if ( !status ) {
-            trace() << "opendir " << nativedir << endl;
-            native_environment = list_native_environment( nativedir );
-            if ( native_environment.empty() )
-                status = 1;
+            char buf[ 1024 ];
+            buf[ 0 ] = '\0';
+            while ( read( pipes[ 0 ], buf, 1023 ) < 0 && errno == EINTR )
+                ;
+            if ( char *nl = strchr( buf, '\n' ))
+                *nl = '\0';
+            native_environment = nativedir + buf;
         }
+        close( pipes[ 0 ] );
         trace() << "native_environment " << native_environment << endl;
-        if ( status ) {
-            rmdir( nativedir.c_str() );
-            return 0;
-        }
-        else {
-            return sumup_dir( nativedir );
+        struct stat st;
+        if ( !status && !native_environment.empty()
+            && stat( native_environment.c_str(), &st ) == 0 ) {
+            return st.st_size;
         }
+        rmdir( nativedir.c_str() );
+        return 0;
     }
     // else
 
-    if ( setgid( nobody_gid ) < 0) {
+#ifndef HAVE_LIBCAP_NG
+    if ( setgid( user_gid ) < 0) {
       log_perror("setgid failed");
       _exit(143);
     }
-    if (!geteuid() && setuid( nobody_uid ) < 0) {
+    if (!geteuid() && setuid( user_uid ) < 0) {
       log_perror("setuid failed");
       _exit (142);
     }
+#endif
 
     if ( chdir( nativedir.c_str() ) ) {
          log_perror( "chdir" );
          _exit(1);
     }
 
-    const char *const argv[] = {
-        BINDIR "/icecc", "--build-native", NULL
-    };
+    close( pipes[ 0 ] );
+    dup2( pipes[ 1 ], 5 ); // icecc-create-env will write the hash there
+    close( pipes[ 1 ] );
+
+    const char ** argv;
+    argv = new const char*[ 4 + extrafiles.size() ];
+    int pos = 0;
+    argv[ pos++ ] = BINDIR "/icecc";
+    argv[ pos++ ] = "--build-native";
+    argv[ pos++ ] = strdup( compiler.c_str());
+    for( list<string>::const_iterator it = extrafiles.begin();
+         it != extrafiles.end(); ++it )
+        argv[ pos++ ] = strdup( it->c_str());
+    argv[ pos++ ] = NULL;
     if ( !exec_and_wait( argv ) ) {
         log_error() << BINDIR "/icecc --build-native failed\n";
         _exit(1);
@@ -278,7 +325,7 @@ size_t setup_env_cache(const string &basedir, string &native_environment, uid_t
 pid_t start_install_environment( const std::string &basename, const std::string &target,
                           const std::string &name, MsgChannel *c,
                           int& pipe_to_stdin, FileChunkMsg*& fmsg,
-                          uid_t nobody_uid, gid_t nobody_gid )
+                          uid_t user_uid, gid_t user_gid )
 {
     if ( !name.size() || name[0] == '.' ) {
         log_error() << "illegal name for environment " << name << endl;
@@ -315,7 +362,7 @@ pid_t start_install_environment( const std::string &basename, const std::string
         return 0;
     }
 
-    if ( chown( dirname.c_str(), 0, nobody_gid ) ||
+    if ( chown( dirname.c_str(), 0, user_gid ) ||
          chmod( dirname.c_str(), 0770 ) ) {
         log_perror( "chown,chmod target" );
         return 0;
@@ -327,7 +374,7 @@ pid_t start_install_environment( const std::string &basename, const std::string
         return 0;
     }
 
-    if ( chown( dirname.c_str(), 0, nobody_gid ) ||
+    if ( chown( dirname.c_str(), 0, user_gid ) ||
          chmod( dirname.c_str(), 0770 ) ) {
         log_perror( "chown,chmod name" );
         return 0;
@@ -348,14 +395,16 @@ pid_t start_install_environment( const std::string &basename, const std::string
         return pid;
     }
     // else
-    if ( setgid( nobody_gid ) < 0) {
+#ifndef HAVE_LIBCAP_NG
+    if ( setgid( user_gid ) < 0) {
       log_perror("setgid fails");
       _exit(143);
     }
-    if (!geteuid() && setuid( nobody_uid ) < 0) {
+    if (!geteuid() && setuid( user_uid ) < 0) {
       log_perror("setuid fails");
       _exit (142);
     }
+#endif
 
     // reset SIGPIPE and SIGCHILD handler so that tar
     // isn't confused when gzip/bzip2 aborts
@@ -384,7 +433,7 @@ pid_t start_install_environment( const std::string &basename, const std::string
 
 
 size_t finalize_install_environment( const std::string &basename, const std::string &target,
-                                     pid_t pid, gid_t nobody_gid)
+                                     pid_t pid, gid_t user_gid)
 {
     int status = 1;
     while ( waitpid( pid, &status, 0) < 0 && errno == EINTR)
@@ -398,7 +447,7 @@ size_t finalize_install_environment( const std::string &basename, const std::str
 
     string dirname = basename + "/target=" + target;
     mkdir( ( dirname + "/tmp" ).c_str(), 01775 );
-    chown( ( dirname + "/tmp" ).c_str(), 0, nobody_gid );
+    chown( ( dirname + "/tmp" ).c_str(), 0, user_gid );
     chmod( ( dirname + "/tmp" ).c_str(), 01775 );
 
     return sumup_dir (dirname);
@@ -434,3 +483,99 @@ size_t remove_environment( const string &basename, const string &env )
 
     _exit(execv(argv[0], argv));
 }
+
+size_t remove_native_environment( const string &env )
+{
+    if ( env.empty() )
+        return 0;
+    struct stat st;
+    if ( stat( env.c_str(), &st ) == 0 ) {
+        unlink( env.c_str());
+        return st.st_size;
+    }
+    return 0;
+}
+
+static void
+error_client( MsgChannel *client, string error )
+{
+    if ( IS_PROTOCOL_23( client ) )
+        client->send_msg( StatusTextMsg( error ) );
+}
+
+void chdir_to_environment( MsgChannel *client, const string &dirname, uid_t user_uid, gid_t user_gid )
+{
+#ifdef HAVE_LIBCAP_NG
+    if ( chdir( dirname.c_str() ) < 0 ) {
+        error_client( client, string( "chdir to " ) + dirname + "failed" );
+        log_perror("chdir() failed" );
+        _exit(145);
+    }
+    if ( chroot( dirname.c_str() ) < 0 ) {
+        error_client( client, string( "chroot " ) + dirname + "failed" );
+        log_perror("chroot() failed" );
+        _exit(144);
+    }
+#else
+    if ( getuid() == 0 ) {
+        // without the chdir, the chroot will escape the
+        // jail right away
+        if ( chdir( dirname.c_str() ) < 0 ) {
+            error_client( client, string( "chdir to " ) + dirname + "failed" );
+            log_perror("chdir() failed" );
+            _exit(145);
+        }
+        if ( chroot( dirname.c_str() ) < 0 ) {
+            error_client( client, string( "chroot " ) + dirname + "failed" );
+            log_perror("chroot() failed" );
+            _exit(144);
+        }
+        if ( setgid( user_gid ) < 0 ) {
+            error_client( client, string( "setgid failed" ));
+            log_perror("setgid() failed" );
+            _exit(143);
+        }
+        if ( setuid( user_uid ) < 0) {
+            error_client( client, string( "setuid failed" ));
+            log_perror("setuid() failed" );
+            _exit(142);
+        }
+    }
+    else {
+        if ( chdir( dirname.c_str() ) ) {
+            log_perror( "chdir" );
+        } else {
+            trace() << "chdir to " << dirname << endl;
+        }
+    }
+#endif
+}
+
+// Verify that the environment works by simply running the bundled bin/true.
+bool verify_env( MsgChannel *client, const string &basedir, const string& target, const string &env,
+                 uid_t user_uid, gid_t user_gid )
+{
+    if ( target.empty() || env.empty())
+        return false;
+
+    string dirname = basedir + "/target=" + target + "/" + env;
+    if ( ::access( string( dirname + "/bin/true" ).c_str(), X_OK ) ) {
+        error_client( client, dirname + "/bin/true is not executable" );
+        log_error() << "I don't have environment " << env << "(" << target << ") to verify." << endl;
+        return false;
+    }
+
+    flush_debug();
+    pid_t pid = fork();
+    assert(pid >= 0);
+    if ( pid > 0) { // parent
+        int status;
+        while ( waitpid( pid, &status, 0 ) < 0 && errno == EINTR )
+            ;
+        return WIFEXITED(status) && WEXITSTATUS(status) == 0;
+    }
+    // child
+    reset_debug(0);
+    chdir_to_environment( client, dirname, user_uid, user_gid );
+    _exit(execl("bin/true", "bin/true", (void*)NULL));
+}
