@@ -26,6 +26,7 @@
 #define _GNU_SOURCE 1
 #endif
 
+#include <sys/stat.h>
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
@@ -46,6 +47,7 @@
 #include <fstream>
 #include <string>
 #include <stdio.h>
+#include <pwd.h>
 #include "comm.h"
 #include "logging.h"
 #include "job.h"
@@ -156,10 +158,11 @@ public:
   int max_jobs;
   bool noremote;
   list<Job*> joblist;
+  int submitted_jobs_count;
   Environments compiler_versions;  // Available compilers
   CS (int fd, struct sockaddr *_addr, socklen_t _len, bool text_based)
     : MsgChannel(fd, _addr, _len, text_based),
-      load(1000), max_jobs(0), noremote(false),
+      load(1000), max_jobs(0), noremote(false), submitted_jobs_count(0),
       state(CONNECTED), type(UNKNOWN), chroot_possible(false)
   {
     hostid = 0;
@@ -179,6 +182,7 @@ public:
   bool chroot_possible;
   static unsigned int hostid_counter;
   map<int, int> client_map; // map client ID for daemon to our IDs
+  map<CS*, Environments> blacklist;
   bool is_eligible( const Job *job );
   bool check_remote( const Job *job ) const;
 };
@@ -195,7 +199,7 @@ class Job
 public:
   unsigned int id;
   unsigned int local_client_id;
-  enum {PENDING, WAITINGFORCS, COMPILING, WAITINGFORDONE} state;
+  enum {PENDING, WAITINGFORCS, COMPILING} state;
   CS *server;  // on which server we build
   CS *submitter;  // who submitted us
   Environments environments;
@@ -215,15 +219,21 @@ public:
   unsigned int arg_flags;
   string language; // for debugging
   string preferred_host; // for debugging daemons
+  bool ignore_unverified; // ignore CSs that don't know M_VERIFY_ENV
   Job (unsigned int _id, CS *subm)
     : id(_id), local_client_id( 0 ), state(PENDING), server(0),
       submitter(subm),
-      starttime(0), start_on_scheduler(0), done_time( 0 ), arg_flags( 0 ) {}
+      starttime(0), start_on_scheduler(0), done_time( 0 ), arg_flags( 0 )
+  {
+    ++submitter->submitted_jobs_count;
+  }
+
   ~Job()
   {
    // XXX is this really deleted on all other paths?
 /*    fd2chan.erase (channel->fd);
     delete channel;*/
+    --submitter->submitted_jobs_count;
   }
 };
 
@@ -377,12 +387,19 @@ server_speed (CS *cs, Job *job)
 
       // we only care for the load if we're about to add a job to it
       if (job) {
-        /* The submitter of a job gets more speed.  So if he is equally
-           fast to the rest of the farm it will be prefered to chose him
-           to compile the job.  Then this can be done locally without
-           needing the preprocessor.  */
-        if (job->submitter == cs)
-          f *= 1.1;
+        if (job->submitter == cs) {
+        /* The submitter of a job gets more speed if it's capable of handling its requests on its own.
+           So if he is equally fast to the rest of the farm it will be prefered to chose him
+           to compile the job.  Then this can be done locally without needing the preprocessor.
+           However if there are more requests than the number of jobs the submitter can handle,
+           it is assumed the submitter is doing a massively parallel build, in which case it is
+           better not to build on the submitter and let it do other work (such as preprocessing
+           output for other nodes) that can be done only locally.  */
+          if (cs->submitted_jobs_count <= cs->max_jobs)
+            f *= 1.1;
+          else
+            f *= 0.1; // penalize heavily
+        }
         else // ignoring load for submitter - assuming the load is our own
           f *= float(1000 - cs->load) / 1000;
       }
@@ -516,6 +533,7 @@ handle_cs_request (MsgChannel *c, Msg *_m)
       job->filename = m->filename;
       job->local_client_id = m->client_id;
       job->preferred_host = m->preferred_host;
+      job->ignore_unverified = m->ignore_unverified;
       enqueue_job_request (job);
       std::ostream& dbg = log_info();
       dbg << "NEW " << job->id << " client="
@@ -644,6 +662,13 @@ envs_match( CS* cs, const Job *job )
   return string();
 }
 
+static bool
+blacklisted( CS* cs, const Job *job, const pair<string, string >& environment )
+{
+  list< pair< string, string > > &blacklist = job->submitter->blacklist[ cs ];
+  return find(blacklist.begin(), blacklist.end(), environment ) != blacklist.end();
+}
+
 /* Given a candidate CS and a JOB, check if any of the requested
    environments could be installed on the CS.  This is the case if that
    env can be run there, i.e. if the host platforms of the CS and of the
@@ -664,7 +689,7 @@ can_install( CS* cs, const Job *job )
 
   for ( Environments::const_iterator it = job->environments.begin(); it != job->environments.end(); ++it )
     {
-      if ( platforms_compatible( it->first, cs->host_platform ) )
+      if ( platforms_compatible( it->first, cs->host_platform ) && !blacklisted( cs, job, *it ))
         return it->first;
     }
   return string();
@@ -680,9 +705,11 @@ bool CS::is_eligible( const Job *job )
 {
   bool jobs_okay = int( joblist.size() ) < max_jobs;
   bool load_okay = load < 1000;
+  bool ignore = job->ignore_unverified && !IS_PROTOCOL_31(this);
   return jobs_okay
     && chroot_possible
     && load_okay
+    && !ignore
     && can_install( this, job ).size()
     && this->check_remote( job );
 }
@@ -1402,6 +1429,22 @@ handle_stats (CS * c, Msg * _m)
   return false;
 }
 
+static bool
+handle_blacklist_host_env(CS * c, Msg * _m)
+{
+  BlacklistHostEnvMsg *m = dynamic_cast<BlacklistHostEnvMsg *>(_m);
+  if (!m)
+    return false;
+  for (list<CS*>::const_iterator it = css.begin(); it != css.end(); ++it)
+    if ( (*it)->name == m->hostname )
+      {
+        trace() << "Blacklisting host " << m->hostname << " for environment " << m->environment
+             << " (" << m->target << ")" << endl;
+        c->blacklist[ *it ].push_back( make_pair( m->target, m->environment ));
+      }
+  return true;
+}
+
 static string
 dump_job (Job *job)
 {
@@ -1412,7 +1455,6 @@ dump_job (Job *job)
 	   job->state == Job::PENDING ? "PEND"
 	     : job->state == Job::WAITINGFORCS ? "WAIT"
 	     : job->state == Job::COMPILING ? "COMP"
-             : job->state == Job::WAITINGFORDONE ? "DONE"
 	     : "Huh?",
 	   job->submitter ? job->submitter->nodename.c_str() : "<>",
 	   job->server ? job->server->nodename.c_str() : "<unknown>");
@@ -1547,7 +1589,7 @@ handle_line (CS *c, Msg *_m)
     }
   else if (cmd == "crashme")
     {
-      *(int *)0 = 42;  // ;-)
+      *(volatile int *)0 = 42;  // ;-)
     }
   else if (cmd == "internals" )
     {
@@ -1702,6 +1744,8 @@ handle_end (CS *toremove, Msg *m)
               ++mit;
             }
         }
+      for( list<CS*>::iterator it = css.begin(); it != css.end(); ++it )
+        (*it)->blacklist.erase( toremove );
     }
     break;
   case CS::LINE:
@@ -1750,6 +1794,7 @@ handle_activity (CS *c)
     case M_LOGIN: ret = handle_relogin (c, m); break;
     case M_TEXT: ret = handle_line (c, m); break;
     case M_GET_CS: ret = handle_cs_request (c, m); break;
+    case M_BLACKLIST_HOST_ENV: ret = handle_blacklist_host_env(c, m); break;
     default:
       log_info() << "Invalid message type arrived " << ( char )m->type << endl;
       handle_end (c, m);
@@ -1834,13 +1879,14 @@ usage(const char* reason = 0)
      cerr << reason << endl;
 
   cerr << "ICECREAM scheduler " VERSION "\n";
-  cerr << "usage: scheduler [options] \n"
+  cerr << "usage: icecc-scheduler [options] \n"
        << "Options:\n"
        << "  -n, --netname <name>\n"
        << "  -p, --port <port>\n"
        << "  -h, --help\n"
        << "  -l, --log-file <file>\n"
        << "  -d, --daemonize\n"
+       << "  -u, --user-uid\n"
        << "  -v[v[v]]]\n"
        << endl;
 
@@ -1873,6 +1919,21 @@ main (int argc, char * argv[])
   bool detach = false;
   int debug_level = Error;
   string logfile;
+  uid_t user_uid;
+  gid_t user_gid;
+
+  struct passwd *pw = getpwnam("icecc");
+  if (pw)
+    {
+      user_uid = pw->pw_uid;
+      user_gid = pw->pw_gid;
+     }
+  else
+    {
+      log_perror ("Error: no icecc user on system. Falling back to nobody.");
+      user_uid = 65534;
+      user_gid = 65533;
+     }
 
   while ( true ) {
     int option_index = 0;
@@ -1882,10 +1943,11 @@ main (int argc, char * argv[])
       { "port", 0, NULL, 'p' },
       { "daemonize", 0, NULL, 'd'},
       { "log-file", 1, NULL, 'l'},
+      { "user-uid", 1, NULL, 'u'},
       { 0, 0, 0, 0 }
     };
 
-    const int c = getopt_long( argc, argv, "n:p:hl:vdr", long_options, &option_index );
+    const int c = getopt_long( argc, argv, "n:p:hl:vdr:u:", long_options, &option_index );
     if ( c == -1 ) break; // eoo
 
     switch ( c ) {
@@ -1928,14 +1990,58 @@ main (int argc, char * argv[])
       else
         usage("Error: -p requires argument");
       break;
+    case 'u':
+        if ( optarg && *optarg )
+          {
+            pw = getpwnam( optarg );
+            if ( !pw )
+              {
+                usage( "Error: -u requires a valid username" );
+               }
+            else
+              {
+                user_uid = pw->pw_uid;
+                user_gid = pw->pw_gid;
+                if ( !user_gid || !user_uid )
+                    usage( "Error: -u <username> must not be root" );
+               }
+           }
+        else
+          {
+            usage( "Error: -u requires a valid username" );
+           }
+        break;
 
     default:
       usage();
     }
   }
 
-  if ( !logfile.size() && detach )
-    logfile = "/var/log/icecc_scheduler";
+  if ( getuid() == 0 )
+    {
+      if ( !logfile.size() && detach ) {
+        if (mkdir("/var/log/icecc", S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH)) {
+          if (errno == EEXIST) {
+              chmod("/var/log/icecc", S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH);
+              chown("/var/log/icecc", user_uid, user_gid);
+          }
+        }
+
+        logfile = "/var/log/icecc/scheduler.log";
+      }
+
+      if ( setgid( user_gid ) < 0 )
+        {
+          log_perror("setgid() failed" );
+          return 1;
+         }
+
+      if ( setuid( user_uid ) < 0 )
+        {
+          log_perror("setuid() failed" );
+          return 1;
+         }
+     }
 
   setup_debug( debug_level, logfile );
   if ( detach )
